/*
  UDP Session 
  by Joel R. Voss
  Jan 20, 2006
  Original Concept by Winston Williams, others
 */

/* Note: 
Send is always on raddr because it is known to be correct.
Poke is the one exception because poke actually finds out the correct port.
Recv is always on oaddr because it can be falsified.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <err.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <pthread.h>

#include <errno.h>
extern int errno;

#include <time.h>

#include <ssl_bf2a.h>
#include <keyex1.h>

#include "list_utils.h"
#include "udp_sess1.h"

const char *udp_control_str(UDP_CONTROL c)
{
    switch(c)
    {
    case UDP_UNKNOWN:
        return "UDP Unknown";
        break;
    case UDP_QUIT:
        return "UDP Quit";
        break;
    case UDP_SEND:
        return "UDP Send";
        break;
    case UDP_SENDTO:
        return "UDP Send To";
        break;
    case UDP_RECV:
        return "UDP Receive";
        break;
    case UDP_RECVFROM:
        return "UDP Receive From";
        break;
    case UDP_ERROR:
        return "UDP Error";
        break;
    case UDP_POKE:
        return "UDP Poke";
        break;
    case UDP_TIMEOUT:
        return "UDP Timeout";
        break;
    default:
        break;
    }
    return "UDP Strange";
}

static void sigint_return(int signo)
{
	fprintf(stderr, "interrupt!\n");
	if(canjump2 == 0)
		return;
	siglongjmp(jmpbuf2, 1);
	return;
}

UDP_CONTROL send_data(char *data, int *data_size, int s, struct sockaddr *raddr)
{
    UDP_CONTROL c = UDP_SENDTO;
    sendto(s, data, *data_size, 0, raddr, sizeof(struct sockaddr_in));
    return c;
}

void *recv_thread(void *arg)
{
	recv_thread_input *conn = (recv_thread_input *)arg;
	int32_t i = 0xffffff00;
	while(1)
	{
		int recv_size;
		char recv_data[1024+52];
		struct sockaddr_in oaddr;
		socklen_t addr_size = sizeof(struct sockaddr_in);
		recv_size = recvfrom(conn->s, recv_data, 1024+52, 0, (struct sockaddr *)&oaddr, &addr_size);
        
		int32_t packet_number_in = i;
		if(conn->secret_len > 0)
		{
			if(memcmp(recv_data, BF_HEADER, 8) == 0)
			{
				char data[1024+52];
				int data_size;
				decrypt_packet(recv_data, recv_size, conn->secret, conn->secret_len, data, &data_size, &packet_number_in);
				//printf("recv_pn: %li\n", packet_number_in);
				
				// Chat is treated specially since it is time sensitive.
				if((packet_number_in & 0xffffff00) == 0xfffffe00)
				{
					if((data_size == 5) && (memcmp(data, "/quit", 5) == 0))
					{
						if(conn->verbosity > 0) fprintf(conn->info_file, "Got a specific remote quit command.\n");
						fprintf(conn->info_file, "\n");
						pthread_mutex_lock(&conn->mut);
						conn->close = 1;
						pthread_mutex_unlock(&conn->mut);
					}
					// Since we're printing it, it's important that it's zero ended. 
					// decrypt_packet _DOES NOT_ terminate the data with null.
					data[data_size] = 0;
					//fprintf(stdout, "(%i)", data_size);
					fwrite(data, data_size, 1, stdout);
					if((data_size == 5) && (memcmp(data, "/quit", 5) == 0))
					{
						fprintf(conn->info_file, "\n");
					}
				}
			}
		}
		//out[recv_size] = 0;
		//printf("recvfrom(%i): %s\n", recv_size, out);
		// Protect conn->packets.
		pthread_mutex_lock(&conn->mut);
		recv_packet_list_add_value(&(conn->packets), recv_data, recv_size, packet_number_in, &oaddr);
		pthread_mutex_unlock(&conn->mut);
		i++;
	}
	// The above loop will never end. That is a feature, not a bug.
	//recv_packet_list_print(&(conn->packets));
	//recv_packet_list_destroy(&(conn->packets));
	return 0;
}

UDP_CONTROL recv_timeout1(recv_thread_input *conn, recv_packet_list **found, int32_t packet_mask)
{
	// Protect conn->packets.
	pthread_mutex_lock(&conn->mut);
	recv_packet_list *found_test = &(conn->packets);
	// found is a marker of where to start unless it is zero.
	if(*found != 0) found_test = *found;
	while(found_test != 0 && found_test->next != 0)
	{
		found_test = recv_packet_list_search(found_test, packet_mask, 1);
		if(found_test != 0)
		{
			*found = found_test;
			pthread_mutex_unlock(&conn->mut);
			return UDP_RECVFROM;
		}
	}
	pthread_mutex_unlock(&conn->mut);
	// Ensure that we don't return something crazy.
	*found = 0;
	return UDP_TIMEOUT;
}

void use_packet(char *out, int *recv_size, recv_thread_input *conn, recv_packet_list *found, struct sockaddr_in *oaddr)
{
	if(found == 0)
	{
		fprintf(conn->err_file, "use_packet sent us a null found.\n");
		return;
	}
	// Protect conn->packets.
	pthread_mutex_lock(&conn->mut);
	if(out)
		memcpy(out, found->packet, found->packet_size);
	if(recv_size)
		*recv_size = found->packet_size;
	if(oaddr != 0)
		memcpy(oaddr, &(found->oaddr), sizeof(struct sockaddr_in));
	recv_packet_list_remove(found);
	if(found != &conn->packets)
		free(found);
	pthread_mutex_unlock(&conn->mut);
}

UDP_CONTROL recv_timeout1_quick(char *out, int *recv_size, recv_thread_input *conn, struct sockaddr_in *raddr, int32_t packet_mask)
{
	recv_packet_list *found = 0;
	UDP_CONTROL c = recv_timeout1(conn, &found, packet_mask);
	if(c == UDP_RECVFROM)
	{
		use_packet(out, recv_size, conn, found, raddr);
		// TODO: Check to see if it really is a /poke.
	}
	return c;
}

UDP_CONTROL send_poke(struct sockaddr_in *raddr, recv_thread_input *conn, int tries)
{
	// Poke a hole in your firewall first. Not encrypted, maybe it should be.
    char poke_data[] = "/poke";
    int data_size = strlen(poke_data);
    UDP_CONTROL c = UDP_UNKNOWN;
	BYTE buf[1024+52];
	int recv_size = -1;
	int nsent = 0;
	while(c != UDP_POKE && c != UDP_RECVFROM && nsent < tries)
	{
		// Reset to ensure no monkeybusiness.
		recv_size = -1;
		UDP_CONTROL cs = send_data(poke_data, &data_size, conn->s, (struct sockaddr *)raddr);
		if(conn->verbosity > 1) fprintf(conn->info_file, "Poke %s (%i)\n", udp_control_str(cs), data_size);
		c = recv_timeout1_quick(buf, &recv_size, conn, raddr, 0xffffff00);
		if(conn->verbosity > 1) fprintf(conn->info_file, "Recv Poke %s (%i)\n", udp_control_str(c), recv_size);
		if(conn->verbosity > 1) fprintf(conn->info_file, "Actual port: %i\n", ntohs(raddr->sin_port));
		nsent++;
		// Wait 1 second after each timeout to receive the reply /poke
		sleep(1);
	}
	if(nsent >= tries)
	{
		fprintf(conn->err_file, "Timed out after %i tries.\n", nsent);
	}
	if(conn->verbosity > -1) fprintf(conn->info_file, "Actual port: %i\n", ntohs(raddr->sin_port));
	
	return c;
}

BIGNUM *recv_dh_public_key(recv_thread_input *conn)
{
	recv_packet_list *found = 0;
	while(1)
	{
		UDP_CONTROL c = recv_timeout1(conn, &found, 0xffffff00);
		
		if(c == UDP_TIMEOUT || found == 0)
		{
			sleep(1);
			return 0;
		}
		// TODO: verify header. re-receive.
		if(found->packet_size <= 0)
		{
			//perror("recv dh");
			//fprintf(conn->info_file, "received dh %s (%i)\n", udp_control_str(c), recv_size);
			//sleep(1);
			//return 0;
			continue;
		}
		
		if(memcmp(found->packet, "dhkeyex_", 8) != 0)
		{
			//fprintf(conn->info_file, "bad dh header %s (%i)\n", udp_control_str(c), found->packet_size);
			//sleep(1);
			//return 0;
			continue;
		}
		
		BIGNUM *pub_key;
		/* Converting from binary to BIGNUM */
		pub_key = BN_bin2bn(found->packet + 8, found->packet_size - 8, NULL);
		use_packet(0, 0, conn, found, 0);
		return pub_key;
	}
}

recv_packet_list *pass_auth_recv(recv_thread_input *conn, char *data, int *data_size, int32_t *packet_number)
{
	recv_packet_list *found = 0;
	while(1)
	{
		// Check for password auth packet as the first packet.
		// Receive
		// wait 1 second for key.
		UDP_CONTROL c = recv_timeout1(conn, &found, 0xffffff00);
		// Check for Salted__ before you decrypt
		if(c == UDP_TIMEOUT || found == 0)
		{
			sleep(1);
			return 0;
		}
		if(conn->verbosity > 1) fprintf(conn->info_file, "Recv %s (%i)\n", udp_control_str(c), found->packet_size);
		if(memcmp(found->packet, BF_HEADER, 8) != 0)
		{
			//fprintf(conn->info_file, "Warning: Recv password header %s (%i)\n", udp_control_str(c), data_size);
			continue;
		}
		decrypt_packet(found->packet, found->packet_size, conn->secret, conn->secret_len, data, data_size, packet_number);
		// Ensure that the data is the correct size.
		if(*data_size != KEY_LENGTH + 16)
		{
			//fprintf(conn->info_file, "Warning: Recv password length %s (%i)\n", udp_control_str(c), data_size);
			continue;
		}
		// Check the header.
		if(memcmp(data, "passauth", 8) != 0)
		{
			//fprintf(conn->info_file, "Warning: Recv password header %8.8s\n", data);
			continue;
		}
		return found;
	}
}

void send_public_key(BIGNUM *pub_key, int s, struct sockaddr_in *raddr)
{
	/* Converting from BIGNUM to binary */
	int len = BN_num_bytes(pub_key);
	BYTE *buf = (BYTE *)malloc(len + 8);
	memcpy(buf, "dhkeyex_", 8);
	len = BN_bn2bin(pub_key, buf + 8) + 8;
	
	// TODO: add header, re-send.
	//UDP_CONTROL c = 
	send_data(buf, &len, s, (struct sockaddr *)raddr);
	//return c;
	free(buf);
}

int dh_keyex(struct sockaddr_in *raddr, struct sockaddr_in *oaddr, recv_thread_input *conn, int tries)
{
    // Secret Key Exchange using Diffie-Hellman
	
	BIGNUM *dh_pub_key = 0;
	BIGNUM *other_dh_pub_key = 0;
	int nsent = 0;
    DH *dh;
    // Generate a key pair
    int ret = dh_gen_key_pair(&dh_pub_key, &dh);
	if(ret != 1) return 1;
	// Try N times (default 100).
	while(other_dh_pub_key == 0 && nsent < tries)
	{
		// Send the public key.
		send_public_key(dh_pub_key, conn->s, raddr);
		// Receive the peer's public key.
		other_dh_pub_key = recv_dh_public_key(conn);
		nsent++;
	}
	if(other_dh_pub_key == 0) return 2;
	// Send the public key again just in case they didn't get it yet.
	send_public_key(dh_pub_key, conn->s, raddr);
    // Create the secret key with this data.
	// secret_len is a i/o variable.
	// secret_len is ususally 256.
	conn->secret_len = 256;
    ret = dh_gen_secret(dh, other_dh_pub_key, conn->secret, &conn->secret_len);
    
    BN_free(other_dh_pub_key);
    BN_free(dh_pub_key);

    return ret;
}

int authenticate(const char *password, struct sockaddr_in *raddr, struct sockaddr_in *oaddr, UDP_AUTH auth, recv_thread_input *conn, int tries)
{
	if(auth == UDP_AUTH_PASSWORD)
	{
		// Password Auth.
		time_t utc_time_data = time(0);
		// time_t is a long, 32-bits.
		BYTE key[32];
		BYTE salt[8];
		// Throwaway variable.
		BYTE iv[EVP_MAX_IV_LENGTH];
		// First 4 bytes of salt are time. Next 4 bytes are random.
		int32_t network_utc_time_data = htonl(utc_time_data);
		memcpy(salt, &network_utc_time_data, 4);
		select_random_salt(salt + 4, 4);
		generate_key(key, password, salt, iv, KEY_LENGTH);
        char data[1024 + 52];
		int data_size = KEY_LENGTH + 16;
		memcpy(data, "passauth", 8);
		memcpy(data+8, salt, 8);
		memcpy(data+16, key, KEY_LENGTH);
		// Send the salt and the key.
        char enc_data[1024 + 52];
        char dec_data[1024];
		int dec_size = -1;
        int enc_size = -1;
		int32_t packet_number = -2;
		int32_t recv_packet_number = -2;
        encrypt_packet(data, data_size, conn->secret, conn->secret_len, enc_data, &enc_size, packet_number);
        int authenticated = 0;
		int nsent = 0;
		while(authenticated == 0 && nsent < tries)
		{
			UDP_CONTROL c = send_data(enc_data, &enc_size, conn->s, (struct sockaddr *)raddr);
			if(conn->verbosity > 1) fprintf(conn->info_file, "Sent %s (%i, %i)\n", udp_control_str(c), enc_size, data_size);

			recv_packet_list *found = pass_auth_recv(conn, dec_data, &dec_size, &recv_packet_number);
			if(found != 0)
			{
				use_packet(0, 0, conn, found, 0);
				authenticated = 1;
			}
		}
		if(authenticated == 0)
		{
			return 1;
		}
		BYTE key_recv[32];
		BYTE *salt_recv = dec_data + 8;
		generate_key(key_recv, password, salt_recv, iv, KEY_LENGTH);
		if(memcmp(dec_data + 16, key_recv, KEY_LENGTH) != 0)
		{
			fprintf(conn->err_file, "Error: Recv password hash\n");
			return 1;
		}
		
		if(memcmp(salt, salt_recv, 8) == 0)
		{
			fprintf(conn->err_file, "Error: Salt is the same, statistically improbably, attack likely.\n");
			return 1;
		}
		
		// Check the date to ensure that it is within acceptable values.
		int32_t *network_recv_time_data = (int32_t *)salt_recv;
		time_t recv_time_data = ntohl(*network_recv_time_data);
		int32_t diff = utc_time_data - recv_time_data;
		if(diff < 0) diff *= -1;
		if(diff > 100)
		{
			fprintf(conn->err_file, "Error: Recv password time is too far off: %i secs. %08x %08x\n", diff, utc_time_data, recv_time_data);
			return 1;
		}
		if(conn->verbosity > -1) fprintf(conn->info_file, "Password authentication succeeded.\n");
	}
	else if(auth == UDP_AUTH_DSA)
	{
		// DSA Auth.
		// TODO: Check for dsa auth packet as the first packet.
	}
	else
	{
		// Null Auth.
		// TODO: Check for null auth packet as the first packet.
	}
    return 0;
}

#ifdef _UDP_CLI

void err1(FILE *err_file, int value, const char *data);
//int safe_read_end();
int usage();

int chat_main(recv_thread_input *conn, struct sockaddr_in *saddr, struct sockaddr_in *raddr)
{
	// chat main only handles sending.
	// recv_thread handles receiving.
	
	int bEndChat = 0;
	int32_t packet_number_out = 0;
    // Chat
    while(1)
    {
        // Send
        int data_size = -1;
		// It needs to be +4 because we're going to add 4 byte packet_number to the end.
        char data[1024+4];
        fgets(data, 1024, stdin);
        // fgets always puts a \0 after the last char.
        // fgets never reads more than size.
        data_size = strlen(data);
        // The exit command is /quit.
        if(strncmp(data, "/quit", 6) == 0)
        {
            if(conn->verbosity > 1) fprintf(conn->info_file, "Got a specific quit command, sending to peer.\n");
            bEndChat = 1;
        }
        if(feof(stdin))
        {
            if(conn->verbosity > 1) fprintf(conn->info_file, "Got an EOF, sending /quit to peer.\n");
            strncpy(data, "/quit", 1024+4);
			data_size = 5;
            bEndChat = 1;
        }
        char enc_data[1024 + 52];
        int enc_size = -1;
        encrypt_packet(data, data_size, conn->secret, conn->secret_len, enc_data, &enc_size, (packet_number_out & 0x000000ff) | 0xfffffe00);
        UDP_CONTROL c = send_data(enc_data, &enc_size, conn->s, (struct sockaddr *)raddr);
        if(conn->verbosity > 2) fprintf(conn->info_file, "Sent %s (%i, %i)\n", udp_control_str(c), enc_size, data_size);
		packet_number_out++;

        // If user initiated an end of chat (EOF or /quit), 
        // quit after sending it.        
        if(bEndChat || conn->close == 1)
        {
            break;
        }
        if(data_size <= 0 || c != UDP_SENDTO)
        {
            fprintf(conn->err_file, "Error: Sent %s (%i)\n", udp_control_str(c), enc_size);
            perror("sendto");
            break;
        }
    }
    return 0;
}

int resend_missing(recv_thread_input *conn, FILE *data_io, struct sockaddr_in *saddr, struct sockaddr_in *raddr, int32_t packet_part, BYTE num_missing, BYTE *missing)
{
	if(conn->verbosity > 1) fprintf(conn->info_file, "%i packets missing from %i:\n", num_missing, packet_part);
	// If it is complete, return to send_main.
	if(num_missing == 0) return 0;
	
	long packet_start = ftell(data_io);
	
	int i;
	int prev_missing = 0;
	for(i = 0; i < num_missing; i++)
	{
		if(fseek(data_io, (prev_missing * -1024) + (missing[i] * 1024), SEEK_CUR) != 0)
			perror("fseek missing packet");
		// The extra 1 accounts for us freading 1024 into the file.
		prev_missing = missing[i] + 1;
		int32_t packet_number = (packet_part & 0xffffff00) | missing[i];
		// get data
        int data_size = -1;
        char data[1024+4];
        data_size = fread(data, 1, 1024, data_io);
		// send data
        int enc_size = -1;
        char enc_data[1024 + 52];
        encrypt_packet(data, data_size, conn->secret, conn->secret_len, enc_data, &enc_size, packet_number);
        UDP_CONTROL c = send_data(enc_data, &enc_size, conn->s, (struct sockaddr *)raddr);
        if(conn->verbosity > 0) fprintf(conn->info_file, "Sending (%i, %i)\n", enc_size, packet_number);

		// If the user interrupts the conn, exit.
		if(conn->close == 1)
 			break;
		// This variable sleep should attempt to maximize speed (decrease sleep) 
		// and minimize packet loss (increase sleep).
		usleep(conn->tspeed * 10);
        if(data_size < 0 || c != UDP_SENDTO)
        {
            fprintf(conn->err_file, "Sent %s (%i)\n", udp_control_str(c), data_size);
            break;
        }
	}

	// Goto the last packet
	if(fseek(data_io, (prev_missing * -1024) + (256 * 1024), SEEK_CUR) != 0)
		perror("fseek last packet");
	// recv missing list
	while(1)
	{
		// send last packet
		int32_t packet_number = packet_part;
		// get data
        int data_size = -1;
        char data[1024+4];
        data_size = fread(data, 1, 1024, data_io);
		// send data
        int enc_size = -1;
        char enc_data[1024 + 52];
        encrypt_packet(data, data_size, conn->secret, conn->secret_len, enc_data, &enc_size, packet_number);
        UDP_CONTROL c = send_data(enc_data, &enc_size, conn->s, (struct sockaddr *)raddr);
        if(conn->verbosity > 0) fprintf(conn->info_file, "Sending (%i, %i)\n", enc_size, packet_number);
		
		// If the user interrupts the conn, exit.
		if(conn->close == 1)
 			break;
		usleep(conn->tspeed * 10);
		
		struct sockaddr_in oaddr;
		int recv_size = -1;
		// get the most recent one
		int found = 0;
		do {
			c = recv_timeout1_quick(enc_data, &recv_size, conn, &oaddr, packet_number & 0xffffff00);
			if(recv_size > 0 && c == UDP_RECVFROM) found = 1;
		} while(c != UDP_TIMEOUT);
		if(found) c = UDP_RECVFROM;
		if(recv_size <= 0 || c != UDP_RECVFROM)
		{
			//fprintf(conn->info_file, "Received %s (%li, %i)\n", udp_control_str(c), packet_number, recv_size);
			if(fseek(data_io, -data_size, SEEK_CUR) != 0)
				perror("fseek -data_size");
			continue;
		}
		if(conn->verbosity > 0) fprintf(conn->info_file, "Received(%i)\n", recv_size);
		int32_t packet_number_in = -1;
		int dec_size = -1;
		char dec_data[1024 + 4];
		// TODO: Check for Salted__ before you decrypt?
		int ret = decrypt_packet(enc_data, recv_size, conn->secret, conn->secret_len, dec_data, &dec_size, &packet_number_in);
		if(ret == 0)
		{
			BYTE *num_missing2 = (BYTE *)dec_data;
			BYTE *missing2 = (BYTE *)dec_data+1;
			if(dec_size < *num_missing2 + 1)
			{
				fprintf(conn->err_file, "Error: Not enough bytes: %i < %i.\n", dec_size, *num_missing2 + 1);
			}
			else
			{
				// Go to the start of the part before you call resend.
				fseek(data_io, packet_start, SEEK_SET);
				return resend_missing(conn, data_io, saddr, raddr, packet_number_in, *num_missing2, missing2);
				// After you resend the packets, you can go onto the next part.
			}
		}
		else
		{
			fprintf(conn->err_file, "Error with received data: %i\n", ret);
		}
	}
	// It should never get here.
	return 1;
}

int send_main(recv_thread_input *conn, FILE *data_io, struct sockaddr_in *saddr, struct sockaddr_in *raddr)
{
	struct sockaddr_in oaddr;

	int32_t packet_number = 0;
	int32_t packets_sent = 0;
	time_t start_time = time(0);
	int32_t packet_start = 0;
	
	// Send file.
    while(1)
    {
        int data_size = -1;
        char data[1024+4];
        data_size = fread(data, 1, 1024, data_io);
		int enc_size = -1;
        char enc_data[1024 + 52];
        encrypt_packet(data, data_size, conn->secret, conn->secret_len, enc_data, &enc_size, packet_number);
        UDP_CONTROL c = send_data(enc_data, &enc_size, conn->s, (struct sockaddr *)raddr);
        if(conn->verbosity > 0) fprintf(conn->info_file, "Sending (%i, %i)\n", enc_size, packet_number);
		packets_sent++;

		// If the user interrupts the conn, exit.
		if(conn->close == 1)
 			break;
		// This variable sleep should attempt to maximize speed (decrease sleep) 
		// and minimize packet loss (increase sleep).
		usleep(conn->tspeed);
        if(data_size < 0 || c != UDP_SENDTO)
        {
            fprintf(conn->err_file, "Sent %s (%i)\n", udp_control_str(c), data_size);
            break;
        }
		// Don't change the packet_number until we have received the resend_list.
		if((packet_number & 0x000000ff) != 0x000000ff)
		{
			packet_number++;
		}
		else
		{
			int recv_size = -1;
			c = recv_timeout1_quick(enc_data, &recv_size, conn, &oaddr, packet_number & 0xffffff00);
			if(recv_size <= 0 || c != UDP_RECVFROM)
			{
				//fprintf(conn->err_file, "Received %s (%li, %i)\n", udp_control_str(c), packet_number, recv_size);
				if(fseek(data_io, -data_size, SEEK_CUR) != 0)
					perror("fseek -data_size");
				continue;
			}
			if(conn->verbosity > 0) fprintf(conn->info_file, "Received(%i)\n", recv_size);
			int32_t packet_number_in = -1;
			int dec_size = -1;
			char dec_data[1024 + 4];
			// TODO: Check for Salted__ before you decrypt?
			int ret = decrypt_packet(enc_data, recv_size, conn->secret, conn->secret_len, dec_data, &dec_size, &packet_number_in);
			if(ret == 0)
			{
				BYTE *num_missing = (BYTE *)dec_data;
				BYTE *missing = (BYTE *)dec_data+1;
				if(dec_size < *num_missing + 1)
				{
					fprintf(conn->err_file, "Error: Not enough bytes: %i < %i.\n", dec_size, *num_missing + 1);
				}
				else
				{
					long packet_end = ftell(data_io);
					// Go to the start of the part before you call resend.
					if(fseek(data_io, packet_start, SEEK_SET))
						perror("fseek packet_start");
					resend_missing(conn, data_io, saddr, raddr, packet_number_in, *num_missing, missing);
					// Go back to the start of the next packet when you're done resending.
					if(fseek(data_io, packet_end, SEEK_SET))
						perror("fseek packet_end");
					packet_start = packet_end;
					// Progress bar.
					print_stats(conn, packet_number, packets_sent, 0, start_time, '\r');
					// After you resend the packets, you can go onto the next part.
					packet_number++;
					
					// If the file has ended, send quit.
					if(feof(data_io) || data_size == 0)
					{
						//fprintf(conn->err_file, "\nbreaking: %li %i %i\n", ftell(data_io), feof(data_io), data_size);
						break;
					}
					/*else
					{
						fprintf(conn->err_file, "\nnot breaking: %li %i %i\n", ftell(data_io), feof(data_io), data_size);
						print_hex(data, 40);
					}*/
				}
			}
			else
			{
				fprintf(conn->err_file, "Error with received data: %i\n", ret);
			}
		}
    }
	print_stats(conn, packet_number-1, packets_sent, 0, start_time, '\n');
    // End the file transfer with /quit.
    // Encrypt the quit of course.
    char data[5+4];
    int data_size = 5; // strlen(data);
    strncpy(data, "/quit", 5+4);
    int enc_size = -1;
    char enc_data[1024 + 52];
	packet_number = 0xfffffe00;
    encrypt_packet(data, data_size, conn->secret, conn->secret_len, enc_data, &enc_size, packet_number);
	data[5] = 0;
    if(conn->verbosity > 0) fprintf(conn->info_file, "Sending %s (%i) for 5 secs.\n", data, enc_size);
	int i;
	for(i = 0; i < 5; i++)
	{
		send_data(enc_data, &enc_size, conn->s, (struct sockaddr *)raddr);
		sleep(1);
	}
    return 0;
}

int recv_main(recv_thread_input *conn, FILE *data_io, struct sockaddr_in *saddr, struct sockaddr_in *raddr)
{
	struct sockaddr_in oaddr;
    
	recv_list part_numbers;
	recv_list_init(&part_numbers, -1, "", 0);
	
	time_t start_time = time(0);
	int32_t part = 0;
	int32_t packets_rcvd = 0;
	
	// Receive file.
    while(1)
    {
		if(conn->close == 1)
 			break;
        char enc_data[1024 + 52];
        int recv_size = -1;
        UDP_CONTROL c = recv_timeout1_quick(enc_data, &recv_size, conn, &oaddr, part);
        if(recv_size < 0 || c != UDP_RECVFROM)
        {
            //fprintf(conn->err_file, "Received %s (%i)\n", udp_control_str(c), recv_size);
            continue;
        }
		int32_t packet_number = -1;
		int data_size = -1;
		char dec_data[1024 + 4];
		// TODO: Check for Salted__ before you decrypt?
		int ret = decrypt_packet(enc_data, recv_size, conn->secret, conn->secret_len, dec_data, &data_size, &packet_number);
        if(conn->verbosity > 1) fprintf(conn->info_file, "Received(%i, %08x)\n", recv_size, packet_number);
		
		// The last packet of the part always repeats until you reply with a repeat list.
		// If they haven't received it, they'll keep sending it, so we'll keep getting it.
		// Perfect.
		if((packet_number & 0x000000ff) == 0x000000ff)
		{
			int32_t min, max;
			recv_list missing;
			recv_list_check(&part_numbers, &min, &max, &missing);
			int reply_size = recv_list_length(&missing) + 1;
			if(conn->verbosity > 0) fprintf(conn->info_file, "Missing (%i, %i, %i)\n", reply_size-1, min, max);
			
			BYTE reply_data[reply_size+4];
			int i;
			// The first byte is the number missing.
			reply_data[0] = reply_size - 1;
			// After that are the packet numbers that are missing.
			recv_list *current_missing = missing.next_recv;
			for(i = 1; i < reply_size; i++)
			{
				reply_data[i] = current_missing->packet_number & 0x000000ff;
				current_missing = current_missing->next_recv;
			}
			
			// FIXME: this is only sent once when missed = 0;
			// Send it to the sender in hopes that you will get what you need.
			int enc_size = -1;
			encrypt_packet(reply_data, reply_size, conn->secret, conn->secret_len, enc_data, &enc_size, packet_number);
			c = send_data(enc_data, &enc_size, conn->s, (struct sockaddr *)raddr);
			if(conn->verbosity > 0) fprintf(conn->info_file, "Sending (%i, %i)\n", enc_size, packet_number);
			
			// If we've accomplished our part, save it and read other packets.
			if(reply_size == 1)
			{
				part += 256;
				// Progress bar.
				print_stats(conn, part-1, packets_rcvd, &part_numbers, start_time, '\r');
				// this ensures that we don't request previous parts.
				recv_list_write(&part_numbers, data_io);
				recv_list_destroy(&part_numbers);
			}
		}
		
		usleep(conn->tspeed);
		// Write the data to the specified file.
		if(ret == 0)
		{
			recv_list_add_value(&part_numbers, packet_number, dec_data, data_size);
			packets_rcvd++;
		}
		else
		{
			fprintf(conn->err_file, "Error with received data: %i\n", ret);
		}
	}
	print_stats(conn, part-1, packets_rcvd, &part_numbers, start_time, '\n');
	recv_list_destroy(&part_numbers);
    return 0;
}

int auth_system(recv_thread_input *conn, struct sockaddr_in *raddr, UDP_AUTH auth, int tries)
{
    char password[256];
	struct sockaddr_in oaddr;

	int nsent = 0;
	/*if(sigsetjmp(jmpbuf2, 1))
	{
		if (nsent >= 1)
		{
			time_t start_time2 = time(0);
			print_stats(&part_numbers, start_time2);
			return 0;
		}
	}*/
	canjump2 = 1;                /* siglongjmp is now OK */
	nsent++;
    
	if (!seed_prng(4))
    {
        fprintf(conn->err_file, "Error: Unable to seed the PRNG.\n");
        return 1;
    }
    
	if(auth != UDP_AUTH_NULL)
	{
		// Secure password retreival algorithm from enc
		password[0] = 0;
		get_password(password, 256, 1);
		// Null password is failure.
		if(password[0] == 0) return 1;
		password[255] = 0;
	}
	
	// Poke finds their "real" local port.
	send_poke(raddr, conn, tries);
	int ret = dh_keyex(raddr, &oaddr, conn, tries);
	if(ret != 0)
	{
		return ret;
	}

	ret = authenticate(password, raddr, &oaddr, auth, conn, tries);
	if(ret != 0)
	{
		return ret;
	}
	if(conn->verbosity > 0) fprintf(conn->info_file, "Properly Authenticated. Commence communication.\n");
	return 0;
}

int print_stats(recv_thread_input *conn, int32_t last_packet_number, int32_t packets, const recv_list *part_numbers, time_t start_time, char end_line)
{
	time_t end_time = time(0);
	//long min, max, gotten;
	/*recv_list missing;
	recv_list_check(part_numbers, &min, &max, &missing);
	if(missing.next_recv != 0)
	{
		fprintf(conn->info_file, "received packets: ");
		recv_list_print(conn->info_file, part_numbers);
		fprintf(conn->info_file, "\n");
		fprintf(conn->info_file, "missing packets: ");
		recv_list_print(conn->info_file, &missing);
		fprintf(conn->info_file, "\n");
	}
	else
	{
		fprintf(conn->info_file, "Received %li-%li packets successfully.\n", min, max);
	}
	// missed = recv_list_length(&missing);
	recv_list_destroy(&missing);
	// max-min-missed
	gotten = recv_list_length(part_numbers);
	*/
	int32_t gotten = last_packet_number + 1;
	float packet_loss = 100.0f * (1.0f * packets - gotten) / gotten;
	float effective_speed = (1.0f * gotten) / (end_time - start_time);
	if(conn->verbosity > -1)
	{
		fprintf(conn->info_file, "%4.2f kB %4.2f kBps (%4.2f%% resent)%c", 1.0f * gotten, effective_speed, packet_loss, end_line);
		fflush(conn->info_file);
	}
    return 0;
}

void err1(FILE *err_file, int value, const char *data)
{
    fprintf(err_file, "Error: %s;\n", data);
    exit(value);
}

/*
static sigjmp_buf jmpbuf;
static int canjump;

static void recv_alarm(int signo)
{
	if(VERBOSE) fprintf(conn->err_file, "alarm!\n");
	if(canjump == 0)
		return;
	siglongjmp(jmpbuf, 1);
}*/

/*
Fixed by recv thread
// This function may be ugly as hell, but it ensures that person does not 
// accidentally execute things that they said that were not sent.
int safe_read_end()
{
	int nsent = 0;
	void *sigfunc1;
	sigfunc1 = signal(SIGALRM, &recv_alarm);
	if(sigsetjmp(jmpbuf, 1))
	{
		if (nsent >= 1)
		{
			return 0;
		}
	}
	canjump = 1;                // siglongjmp is now OK //
	alarm(1);
	nsent++;
	while(!feof(stdin))
	{
		char buf[24];
		fread(buf, 24, 1, stdin);
	}
	canjump = 0;
	alarm(0);
	signal(SIGALRM, sigfunc1);
	return 0;
}
*/

int usage()
{
    printf("Usage:   udp_sess1 [-c] [-r] [-f file] [-s] [-n] [-d] [-t tries] [-v[v[v]]]\n             local_port remote_addr remote_port\n");
    printf("         -c       : Chat only, no files.\n");
    printf("         -r       : Receive file.\n");
    printf("         -f file  : Read or Save file in filesystem.\n");
    printf("         -n       : Null Authentication.\n");
    printf("         -d       : DSA Authentication.\n");
    printf("         -t tries : Number of pokes to send/recv (connect timeout secs).\n");
    printf("         -v       : Increase verbosity up to -vvv.\n");
    printf("         The default options are to send a file from stdin.\n");
    printf("         All data is encrypt with a password using blowfish-cbc.\n");
    printf("Example: udp_sess1 3111 dmitry.altsci.com 3222 < r.txt\n");
    printf("         udp_sess1 -f d.txt 3111 dmitry.altsci.com 3222\n");
    printf("         udp_sess1 -r 3111 dmitry.altsci.com 3222 > d.txt\n");
    printf("         udp_sess1 -c 3111 dmitry.altsci.com 3222\n");
    printf("         udp_sess1 -c -s 3111 dmitry.altsci.com 3222\n");
    return 1;
}

int main(int argc, char **argv)
{
    struct sockaddr_in raddr;
    //struct sockaddr_in saddr;
    //int s;
    int ret;
    
    int chat = 0;
    int send = 1;
    char *filename = 0;
    int dargs = 0;
    int local_io_port = 0;
    char local_io_addr[] = "0.0.0.0";
    int remote_recv_port = 0;
    char *remote_hostname = 0;
	UDP_AUTH auth = UDP_AUTH_PASSWORD;
	int tries = default_tries;
	int speed = 1000;
	int verbosity = 0;

    if(argc < 4) return usage();
    
    int i;
    for(i = 1; i < argc; i++)
    {
        if(strcmp(argv[i], "-c") == 0)
        {
            chat = 1;
        }
        else if(strcmp(argv[i], "-r") == 0)
        {
            send = 0;
        }
        else if(strcmp(argv[i], "-f") == 0)
        {
            // -f file
            i++;
            filename = argv[i];
        }
        else if(strcmp(argv[i], "-n") == 0)
        {
            // -n Null authentication
			auth = UDP_AUTH_NULL;
        }
        else if(strcmp(argv[i], "-d") == 0)
        {
            // -d DSA authentication
			auth = UDP_AUTH_DSA;
        }
        else if(strcmp(argv[i], "-t") == 0)
        {
            // -t tries
			i++;
			tries = atoi(argv[i]);
        }
        else if(strcmp(argv[i], "-s") == 0)
        {
            // -s speed
			i++;
			speed = atoi(argv[i]);
        }
        else if(strcmp(argv[i], "-v") == 0)
        {
            // -v
			verbosity++;
        }
        else if(strcmp(argv[i], "-vv") == 0)
        {
            // -vv
			verbosity += 2;
        }
        else if(strcmp(argv[i], "-vvv") == 0)
        {
            // -vvv
			verbosity += 3;
        }
        else
        {
            if(dargs == 0)
            {
                //fprintf(conn->err_file, "setting local io port: %s, %i\n", argv[i], atoi(argv[i]));
                local_io_port = atoi(argv[i]);
            }
            if(dargs == 1)
            {
                //fprintf(conn->err_file, "setting remote recv addr: %s\n", argv[i]);
                remote_hostname = argv[i];
            }
            if(dargs == 2)
            {
                //fprintf(conn->err_file, "setting remote recv port: %s, %i\n", argv[i], atoi(argv[i]));
                remote_recv_port = atoi(argv[i]);
            }
            dargs++;
        }
    }
    // It is an error condition to not give all three data args.
    if(dargs < 3)
    {
        return usage();
    }
    
    FILE *data_io;
    if(filename == 0)
    {
        if(send == 1)
        {
            data_io = stdin;
        }
        else
        {
            data_io = stdout;
        }
    }
    else
    {
        if(send == 1)
        {
            data_io = fopen(filename, "rb");
        }
        else
        {
            data_io = fopen(filename, "wb");
        }
    }
	
	if(data_io <= 0)
	{
		err1(stderr, 1, "fopen");
		return 1;
	}
    
	pthread_t recv_tid;
	pthread_attr_t recv_attr;

	recv_thread_input conn;
	
	conn.secret_len = 0;
	conn.close      = 0;
	conn.err_file   = stderr;
	conn.info_file  = stdout; //fopen("testx14s.txt", "wb");
	conn.verbosity  = verbosity;
	// TSpeed is inversely proportional to the input speed.
	// speed 100 = 1000us
	// speed 200 = 500us
	// speed 1   = 100000us
	if(speed > 0)
		conn.tspeed      = 100000 / speed;
	else
		conn.tspeed      = 100000;
	
	conn.s = socket(PF_INET, SOCK_DGRAM, 0);
    if(conn.s < 0) err1(conn.err_file, 1, "socket: SOCK_DGRAM");

    memset(&conn.saddr, 0, sizeof(struct sockaddr_in));
    conn.saddr.sin_family = AF_INET;
    conn.saddr.sin_port = htons(local_io_port);
    conn.saddr.sin_addr.s_addr = inet_addr(local_io_addr);

    
	struct hostent     *hp;
	if ((hp = gethostbyname(remote_hostname)) == 0) { // do we know the host's //
		errno = ECONNREFUSED;                         // address? //
		err1(conn.err_file, 1, "gethostbyname");                     // no //
	}
	memset(&raddr, 0, sizeof(struct sockaddr));
    memcpy((char *)&raddr.sin_addr, hp->h_addr, hp->h_length);     // set address //
    raddr.sin_family = hp->h_addrtype;
    //raddr.sin_family = AF_INET;
    //raddr.sin_addr.s_addr = inet_addr(remote_recv_addr);
    raddr.sin_port = htons(remote_recv_port);
    
    //fprintf(conn->err_file, "bind(%s, %i)\n", local_io_addr, local_io_port);
    ret = bind(conn.s, (struct sockaddr *)&conn.saddr, sizeof(struct sockaddr_in));
    if(ret) err1(conn.err_file, 1, "bind");

    if(conn.verbosity > -1) fprintf(conn.info_file, "Connecting to %s (%s):%i\n", remote_hostname, inet_ntoa(raddr.sin_addr), remote_recv_port);
    
	recv_packet_list_init(&(conn.packets), "", 0, -1, &raddr);
	
	// Setup the mutex that protects conn.packets during threading.
	pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;
	memcpy (&conn.mut, &mut, sizeof(pthread_mutex_t));
    
	// The receive thread is a loop that receives all the time, no waiting.
	// It turns every recv_data into a recv_timeout1
	pthread_attr_init(&recv_attr);
	pthread_attr_setdetachstate(&recv_attr,PTHREAD_CREATE_DETACHED);
	pthread_attr_setscope(&recv_attr, PTHREAD_SCOPE_SYSTEM);
	pthread_create(&recv_tid, &recv_attr, recv_thread, (void *)&conn);
	
	// Catch interrupts, send them to this simple function.
	//void *sigfunc1;
	//sigfunc1 = signal(SIGINT, &sigint_return);
	
	if(auth_system(&conn, &raddr, auth, tries) == 0)
	{
		
		if(chat == 1)
		{
			ret = chat_main(&conn, &(conn.saddr), &raddr);
			// Fixed by recv thread: Ensure that they don't execute accidentally!
			//safe_read_end();
		}
		else if(send == 1)
		{
			ret = send_main(&conn, data_io, &(conn.saddr), &raddr);
		}
		else //if(send == 0)
		{
			ret = recv_main(&conn, data_io, &(conn.saddr), &raddr);
		}
	}
    //signal(SIGINT, sigfunc1);
	
	// Free memory allocated and stored in conn.packets
	pthread_mutex_lock(&conn.mut);
	recv_packet_list_destroy(&(conn.packets));
	pthread_mutex_unlock(&conn.mut);
    
	fclose(data_io);
	fclose(conn.info_file);
	fclose(conn.err_file);
    shutdown(conn.s, SHUT_WR);
    return ret;
}
#endif // _UDP_CLI
